import copy
from .pnet import PNet
import polars as pl
import os

class SimulationHistoryHandler:
    def __init__(self, places, parquet_file, flush_every=10000):
        self.places = places
        self.parquet_file = parquet_file
        self.flush_every = flush_every
        self.rows = []
        self.last_firing = None
        self._parquet_initialized = False

    def __call__(self, dict_places, firing_sequence, step):
        # Determine the last transition fired (None for step 0)
        if step == 0 or not firing_sequence:
            transition = None
        else:
            transition = firing_sequence[-1]
        row = {'transition': transition, 'step': step}
        for place in self.places:
            row[place] = dict_places.get(place, 0)
        self.rows.append(row)
        if len(self.rows) >= self.flush_every:
            self.flush()

    def flush(self):
        if not self.rows:
            return
        df = pl.DataFrame(self.rows)
        if not self._parquet_initialized or not os.path.exists(self.parquet_file):
            df.write_parquet(self.parquet_file)
            self._parquet_initialized = True
        else:
            # Append to existing parquet file
            old_df = pl.read_parquet(self.parquet_file)
            new_df = pl.concat([old_df, df])
            new_df.write_parquet(self.parquet_file)
        self.rows = []

    def finalize(self):
        self.flush()


def simulate_with_history_to_parquet(pnet: PNet, num_steps: int, parquet_file: str, flush_every: int = 10000, law: str = 'random'):
    """
    Runs a simulation and stores the full history to a Parquet file using polars.
    """
    places = list(pnet.dict_places.keys())
    handler = SimulationHistoryHandler(places, parquet_file, flush_every)
    pnet.simulate_petrinet(num_steps, law, handler)
    handler.finalize()
    return parquet_file 


from petri_net_core.petrinet.pnet import PNet
from petri_net_core.object_petrinet.opnet2pnet import ObjectPNet2PNetAdapter
from petri_net_core.object_petrinet.opnet_validator import OPNetValidator
import copy

class ObjectPNet():
    def __init__(self, opnet_dict):
        OPNetValidator.validate_schema(opnet_dict)
        OPNetValidator.validate_integrity(opnet_dict)
        self.opnet_definition=opnet_dict
        self.adapter=ObjectPNet2PNetAdapter(opnet_dict)
        self.pnet_dict=self.adapter.get_pnet_json_dict()
#        print(self.pnet_dict)
        self.pnet=PNet(self.pnet_dict)

    def get_place_tokens(self,pnet_name,place_name):
        adapted_place_name=self.adapter.get_adapted_pnet_place_name(pnet_name,place_name)
        return self.pnet.get_tokens([adapted_place_name])[adapted_place_name]
    
    def get_multiple_place_tokens(self,pnet_place_dict):
        output_dict={}
        for pnet in pnet_place_dict:
            place_array=pnet_place_dict[pnet]
            if len(place_array)>0:
                output_dict[pnet]={}
            for place in place_array:
                output_dict[pnet][place]=self.get_place_tokens(pnet,place)
        return output_dict
    
    def reset(self):
        self.pnet.reset()

    def get_vanilla_pnet_def(self):
        return copy.deepcopy(self.pnet_dict)
        
    def simulate(self,num_steps:int,law:str='random',handler=False):
        self.pnet.simulate_petrinet(num_steps=num_steps,law=law,handler=handler)

    def get_firing_sequence(self):
        return self.pnet.get_firing_sequence()
        


import unittest
import tempfile
import os
import polars as pl
from petri_net_core.petrinet.pnet import PNet
from petri_net_core.petrinet.pnet_simulation_helper import simulate_with_history_to_parquet

pnet_dict={
  "places": {
    "p_A": 10,
    "p_B": 30,
    "p_C": 0,
    "p_D": 0
  },
  "transitions":{
      "t_1":{
        "consume": {
          "p_A": 1,
          "p_B": 2
        },
        "produce": {
          "p_C": 1
        }
      },
      "t_2":{
        "consume": {
          "p_B": 1,
          "p_C": 1
        },
        "produce": {
          "p_D": 2
        }
      }
  }
}

class PNetSimulationHelperTests(unittest.TestCase):
    def setUp(self) -> None:
        self.simple_ok_pnet = pnet_dict
        return super().setUp()

    def test_simulation_history_to_parquet(self):
        OkPnet = PNet(self.simple_ok_pnet)
        with tempfile.TemporaryDirectory() as tmpdir:
            parquet_file = os.path.join(tmpdir, 'history.parquet')
            simulate_with_history_to_parquet(OkPnet, num_steps=5, parquet_file=parquet_file, flush_every=2)
            df = pl.read_parquet(parquet_file)
            # Check columns
            expected_columns = {'transition', 'step', 'p_A', 'p_B', 'p_C', 'p_D'}
            self.assertTrue(expected_columns.issubset(set(df.columns)))
            # Check number of rows (should be 5)
            self.assertEqual(df.shape[0], 5)
            # Check that step column is sequential
            self.assertListEqual(list(df['step']), list(range(5)))
            # Check that the first transition is None
            self.assertIsNone(df['transition'][0])
            # Check that place values are integers
            for place in ['p_A', 'p_B', 'p_C', 'p_D']:
                self.assertTrue(df[place].to_numpy().dtype.kind in {'i', 'u'})

if __name__ == '__main__':
    unittest.main() 


import numpy as np
import copy
from petri_net_core.petrinet.pnet_validator import PNetValidator

rand=np.random

# TODO transformar pnet em uma interface, fazer com que essa classe seja uma vanilla pnet que implementa a interface e ajustar para que a opnet tb implemente a interface

class PNet:
    def __init__(self,pnet_dict:dict):
        PNetValidator.validate_schema(pnet_dict)
        PNetValidator.validate_integrity(pnet_dict)
        dict_places=pnet_dict['places']
        self.initial_dict_places=copy.deepcopy(dict_places)
        self.transitions_dict=pnet_dict['transitions']
        self.reset()
        
    
    def reset(self):
        self.firing_sequence=[]
        self.dict_places=copy.deepcopy(self.initial_dict_places)
        self.transition_count={}
        for transition in self.transitions_dict:
            self.transition_count[transition]=0

    def _initialize_arrays(self):
        self.places=[]
        self.initial_tokens=np.zeros(len(self.dict_places))
        
        j=0
        for i in self.dict_places:
            self.initial_tokens[j]=self.dict_places[i]
            self.places.append(i)
            j+=1

    def get_tokens(self,places_list=[]):
        if not places_list:
            return self.dict_places
        tokens={}
        for place in places_list:
            tokens[place]=self.dict_places[place]
        return tokens
    
    def get_firing_sequence(self):
        return self.firing_sequence

    def add_transition(self,name:str,consume:dict,produce:dict):
        self.transitions_dict[name]={"consume":consume,"produce":produce}
        self.transition_count[name]=0

    def add_multiple_transitions(self,transition_array):
        for transition in transition_array:
            self.add_transition(transition['name'],transition['consume'],transition['produce'])

    def __step(self,transition_name:str, skip_fireability_check:bool=False):
        
        fire=True
        
        if not skip_fireability_check:
            fire=self.check_fireability(transition_name)[0]

        if fire:
            consume_dict=self.transitions_dict[transition_name]["consume"]
            produce_dict=self.transitions_dict[transition_name]["produce"]
            for consumed_place in consume_dict:
                self.dict_places[consumed_place]-=consume_dict[consumed_place]
            for produced_place in produce_dict:
                self.dict_places[produced_place]+=produce_dict[produced_place]
            self.firing_sequence.append(transition_name)
            self.transition_count[transition_name]+=1
            
    def check_fireability(self,transition_name:str)->tuple[bool,int]:
        consume_dict=self.transitions_dict[transition_name]["consume"]
        mass=1
        counter=0
        for consumed_place in consume_dict:
            if consume_dict[consumed_place]>self.dict_places[consumed_place]:
                return False,0
            mass*=self.dict_places[consumed_place]
            counter+=1
        
        try:
            mass_action=mass**(1/counter)
        except TypeError:
            print(transition_name)
            print(f'mass: {mass}')
            print(f'counter: {counter}')    
        
        return True,mass_action

    def random_valid_step(self):
        valid_transitions=[]
        for transition in self.transitions_dict:
            if self.check_fireability(transition)[0]:
                valid_transitions.append(transition)
        if len(valid_transitions)>0:
            random_transition=rand.choice(valid_transitions)
            self.__step(random_transition,True)
        else:
            raise NoMoreValidTransitionsException("No more valid transitions possible")
    
    '''
    step_logic should be a function that receives a 
    '''
    def custom_valid_step(self, step_logic):
        valid_transitions=[]
        for transition in self.transitions_dict:
            if self.check_fireability(transition)[0]:
                valid_transitions.append(transition)
        if len(valid_transitions)>0:
            transition=step_logic(copy.deepcopy(self.dict_places), valid_transitions)
            if transition in valid_transitions:
                self.__step(transition,True)
            else:
                raise InvalidTransitionFiredException(f'Custom step_logic returned an invalid transition: {transition}, expected a transition in: {",".join(valid_transitions)}')
        else:
            raise NoMoreValidTransitionsException("No more valid transitions possible")
        
    def random_mass_action_step(self):
        valid_transitions=[]
        sum_mass_action=0
        for transition in self.transitions_dict:
            valid,mass_action=self.check_fireability(transition)
            if valid:
                valid_transitions.append([transition,mass_action])
                sum_mass_action+=mass_action
        if len(valid_transitions)>0:
            uniform_random=rand.rand()
            cumsum_mass=0
            for transition in valid_transitions:
                cumsum_mass+=(transition[1]/sum_mass_action)
                if uniform_random<=cumsum_mass:
                    random_transition=transition[0]
                    break
                    
            self.__step(random_transition,True)
        else:
            raise NoMoreValidTransitionsException("No more valid transitions possible")
        
    def simulate_petrinet(self,num_steps:int,law:str='random',handler=None):
        self.reset()
        for i in range(num_steps):
            try:
                if law=='random':
                    self.random_valid_step()
                elif law=='mass_action':
                    self.random_mass_action_step()
                else:
                    raise InvalidLawException('Law parameter must be either "random" or "mass_action".')
                if handler is not None:
                    handler(copy.deepcopy(self.dict_places), copy.deepcopy(self.firing_sequence),i)
            except NoMoreValidTransitionsException:
                #print(f"No more valid transitions, ending simulation at step: {i}")
                break

class DuplicateTransition(Exception):
    pass    

class NoMoreValidTransitionsException(Exception):
    pass

class InvalidTransitionFiredException(Exception):
    pass

class InvalidLawException(Exception):
    pass




import unittest
import petri_net_core.object_petrinet.opnet as opnet
import petri_net_core.petrinet.pnet_validator as pnet_validator
from petri_net_core.object_petrinet.opnet_validator import UnknownPlaceException
from jsonschema.exceptions import ValidationError
import copy

##  python -m unittest tests/opnet_tests.py -v

json_input={
    "pnet1":{
      "places": {
        "m": {
            "pnet":"pnet2",
            "tokens":1
        },
        "p": 0
      },
      "transitions": 
        {
          "translation":{
              "consume": {
                "m": 1
              },
              "produce": {
                "p": 1
              }
          }
         }
      
    },
    "pnet2":{
      "places": {
        "m0": 1,
        "m1": 0,
        "m2": 0,
        "a":2
      },
      "transitions": 
        {
            "elongation_step1":{
              "consume": {
                "m0": 1,
                "a":1
              },
              "produce": {
                "m1": 1
              }
              },
            "elongation_step2":{
              "consume": {
                "m1": 1,
                "a":1
              },
              "produce": {
                "m2": 1
              }
              },
              "translation":{
              "consume": {
                "m2": 1
              },
              "produce":{}
            }
        }
    }
}

class OPNetTests(unittest.TestCase):
    def setUp(self) -> None:
        self.opnet=opnet.ObjectPNet(json_input)
        return super().setUp()
    
    
    def test_rejects_invalid_schema_for_subpnet(self):
        invalid_opnet=copy.deepcopy(json_input)
        invalid_opnet['invalid_subpnet']=[]
        with self.assertRaises(ValidationError):
            opnet.ObjectPNet(invalid_opnet)
        del invalid_opnet['invalid_subpnet']
        invalid_opnet['invalid_subpnet']={
            'places':''
        }
        with self.assertRaises(ValidationError):
            opnet.ObjectPNet(invalid_opnet)
            
    def test_rejects_inconsistent_transitions(self):
        invalid_opnet=copy.deepcopy(json_input)
        invalid_opnet['pnet2']['transitions']['translation']['produce']['p']=1
        with self.assertRaises(ValidationError):
            opnet.ObjectPNet(UnknownPlaceException)
        
    def test_generates_valid_petrinet(self):
        pnet_json_dict=self.opnet.get_vanilla_pnet_def()
        pnet_validator.PNetValidator.validate_schema(pnet_json_dict)
        pnet_validator.PNetValidator.validate_integrity(pnet_json_dict)
        
    def test_access_opnet(self):
        self.opnet.reset()
        self.assertEqual(self.opnet.get_place_tokens('pnet1','m'),1)
        self.assertEqual(self.opnet.get_place_tokens('pnet1','p'),0)
        self.assertEqual(self.opnet.get_place_tokens('pnet2','m0'),1)
        self.assertEqual(self.opnet.get_place_tokens('pnet2','m1'),0)
        self.assertEqual(self.opnet.get_place_tokens('pnet2','m2'),0)
        self.assertEqual(self.opnet.get_place_tokens('pnet2','a'),2)

    def test_batch_access_opnet(self):
        self.opnet.reset()
        place_dict={
            'pnet1':['m','p'],
            'pnet2':['m2']
        }
        returned_dict=self.opnet.get_multiple_place_tokens(place_dict)
        self.assertDictEqual(returned_dict,{
            'pnet1':{'m':1,'p':0},
            'pnet2':{'m2':0},
        })
    
    def test_reset_opnet(self):
        self.opnet.simulate(1)
        self.opnet.reset()
        self.assertEqual(self.opnet.get_place_tokens('pnet1','m'),1)
        self.assertEqual(self.opnet.get_place_tokens('pnet1','p'),0)
        self.assertEqual(self.opnet.get_place_tokens('pnet2','m0'),1)
        self.assertEqual(self.opnet.get_place_tokens('pnet2','m1'),0)
        self.assertEqual(self.opnet.get_place_tokens('pnet2','m2'),0)
        self.assertEqual(self.opnet.get_place_tokens('pnet2','a'),2)

    def test_dynamic_opnet(self):        
        self.opnet.simulate(1)
        self.assertEqual(self.opnet.get_place_tokens('pnet2','m0'),0)
        self.assertEqual(self.opnet.get_place_tokens('pnet2','m1'),1)
        self.assertEqual(self.opnet.get_place_tokens('pnet2','a'),1)
        self.opnet.simulate(2)
        self.assertEqual(self.opnet.get_place_tokens('pnet2','m1'),0)
        self.assertEqual(self.opnet.get_place_tokens('pnet2','m2'),1)
        self.assertEqual(self.opnet.get_place_tokens('pnet2','a'),0)
        self.opnet.simulate(3)
        self.assertEqual(self.opnet.get_place_tokens('pnet2','m2'),0)
        self.assertEqual(self.opnet.get_place_tokens('pnet1','m'),0)
        self.assertEqual(self.opnet.get_place_tokens('pnet1','p'),1)
        self.assertListEqual(self.opnet.get_firing_sequence(),['t_elongation_step1', 't_elongation_step2', 't_translation'])
                
    



if __name__ == '__main__':
    unittest.main()


import unittest
from petri_net_core.petrinet.pnet import PNet
from petri_net_core.petrinet.pnet import InvalidTransitionFiredException
import copy
from jsonschema.exceptions import ValidationError
from petri_net_core.petrinet.pnet_validator import UnknownPlaceException
##  python -m unittest tests/pnet_tests.py  -v

## TODO adicionar teste de que tudo continua funcionando mesmo com transições que não possuem places consumidos ou places produzidos

pnet_dict={
  "places": {
    "p_A": 10,
    "p_B": 30,
    "p_C": 0,
    "p_D": 0
  },
  "transitions":{
      "t_1":{
        "consume": {
          "p_A": 1,
          "p_B": 2
        },
        "produce": {
          "p_C": 1
        }
      },
      "t_2":{
        "consume": {
          "p_B": 1,
          "p_C": 1
        },
        "produce": {
          "p_D": 2
        }
      }
  }
}

class PNetTests(unittest.TestCase):
    def setUp(self) -> None:
        self.simple_ok_pnet=pnet_dict
        return super().setUp()

    def test_correct_build_given_valid_input(self):
        OkPnet=PNet(self.simple_ok_pnet)
        self.assertEqual(OkPnet.dict_places['p_A'],10)
        self.assertEqual(OkPnet.dict_places['p_B'],30)
        self.assertEqual(OkPnet.dict_places['p_C'],0)
        self.assertEqual(OkPnet.dict_places['p_D'],0)
        self.assertEqual(OkPnet.transitions_dict['t_1']['consume']["p_A"],1)
        self.assertEqual(OkPnet.transitions_dict['t_1']['consume']["p_B"],2)
        self.assertEqual(OkPnet.transitions_dict['t_1']['produce']["p_C"],1)
        self.assertEqual(OkPnet.transitions_dict['t_2']['consume']["p_B"],1)
        self.assertEqual(OkPnet.transitions_dict['t_2']['consume']["p_C"],1)
        self.assertEqual(OkPnet.transitions_dict['t_2']['produce']["p_D"],2)
    
    def test_rejects_invalid_schema_1(self):
        invalid_pnet=copy.deepcopy(self.simple_ok_pnet)
        del invalid_pnet['places']
        with self.assertRaises(ValidationError):
            PNet(invalid_pnet)
    
    def test_rejects_invalid_schema_2(self):
        invalid_pnet=copy.deepcopy(self.simple_ok_pnet)
        del invalid_pnet['transitions']
        with self.assertRaises(ValidationError):
            PNet(invalid_pnet)
    
    def test_rejects_invalid_schema_3(self):
        invalid_pnet=copy.deepcopy(self.simple_ok_pnet)
        del invalid_pnet['places']['p_B']
        invalid_pnet['places']['p_B+']=30
        with self.assertRaises(ValidationError):
            PNet(invalid_pnet)
    
    def test_rejects_invalid_schema_negative_tokens(self):
        invalid_pnet=copy.deepcopy(self.simple_ok_pnet)
        invalid_pnet['places']['p_B']=-3
        with self.assertRaises(ValidationError):
            PNet(invalid_pnet)

    def test_rejects_invalid_schema_fractional_tokens(self):
        invalid_pnet=copy.deepcopy(self.simple_ok_pnet)
        invalid_pnet['places']['p_B']=3.5
        with self.assertRaises(ValidationError):
            PNet(invalid_pnet)

    def test_rejects_inconsistent_transitions(self):
        invalid_pnet=copy.deepcopy(self.simple_ok_pnet)
        del invalid_pnet['transitions']['t_1'] ['consume']['p_A']
        invalid_pnet['transitions']['t_1'] ['consume']['p_X']=1
        with self.assertRaises(UnknownPlaceException):
            PNet(invalid_pnet)

    def test_runs_steps_correctly(self):
        OkPnet=PNet(self.simple_ok_pnet)
        OkPnet.custom_valid_step(step_transition_t1)
        self.assertEqual(OkPnet.dict_places['p_A'],9)
        self.assertEqual(OkPnet.dict_places['p_B'],28)
        self.assertEqual(OkPnet.dict_places['p_C'],1)

    def test_rejects_invalid_step(self):
        OkPnet=PNet(self.simple_ok_pnet)
        for i in range(10):
          OkPnet.custom_valid_step(step_transition_t1)
        self.assertEqual(OkPnet.dict_places['p_A'],0)
        with self.assertRaises(InvalidTransitionFiredException):
          OkPnet.custom_valid_step(step_transition_t1)  
        
def step_transition_t1(dict_places, valid_transitions):
    return 't_1'

def step_transition_t2(dict_places, valid_transitions):
    return 't_2'

def insecure_step_function(dict_places, valid_transitions):
    dict_places['p_A']=1
    return 't_1'


if __name__ == '__main__':
    unittest.main()


from jsonschema import validate as jsonschemavalidate
import json
import os.path as path
import os

class OPNetValidator():
    def __init__(self):
        pass
    
    @staticmethod
    def validate_schema(opnet_dict):
        schema_path = os.path.join(os.path.dirname(__file__), 'opnet_schema.json')
        with open(schema_path, 'r') as opnet_json_schema_path:
            pnet_json_schema=json.load(opnet_json_schema_path)
            jsonschemavalidate(opnet_dict,schema=pnet_json_schema)

    '''
    Validate that the transitions reference only places declared in their own networks
    '''
    @classmethod
    def validate_integrity(cls, opnet_def:dict):
        for pnet in opnet_def:
            pnet_places=opnet_def[pnet]['places']
            pnet_transitions=opnet_def[pnet]['transitions']
            helpful_error_strings=[]
            for transition_key in pnet_transitions:
                transition=pnet_transitions[transition_key]
                for consumed_place in transition['consume']:
                    if consumed_place not in pnet_places:
                        helpful_error_strings.append(f'Consumed place {consumed_place} in transition {transition_key} not declared in PNet {pnet}') 
                for produced_place in transition['produce']:
                    if produced_place not in pnet_places:
                        helpful_error_strings.append(f'Produced place {produced_place} in transition {transition_key} not declared in PNet {pnet}')            
            if len(helpful_error_strings)>0:
                raise UnknownPlaceException('\n'+'\n'.join(helpful_error_strings))

class UnknownPlaceException(Exception):
    pass
